import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import (MultipleLocator, FormatStrFormatter, AutoMinorLocator)
from radpy.UDfitting import UDV2
from radpy.LDfitting import V2
plt.rcParams['text.usetex'] = True


# Function to bin the PAVO data
def bin_data(x, y, dy, bin_width=5e6, min_points_per_bin=1):
    ###########################################################################
    # Function: bin_data                                                      #
    # Inputs: x -> the x data to bin                                          #
    #         y -> the y data to bin                                          #
    #         bin_width -> the width of the data bins                         #
    #                      default is 5e6 but can be changed                  #
    #         min_points_per_bin-> ensures every bin has 1 data point in it   #
    #                              to avoid nans and issues                   #
    # Outputs: binned_x, binned_y, and binned_dy                              #
    # What it does:                                                           #
    #       1. ensures the input data are arrays                              #
    #       2. sorts the data to make sure its binning properly               #
    #       3. determines the number of bins needed for the data              #
    #       4. sets the bin indices                                           #
    #       5. goes through the data and assigns the data into a bin          #
    #       6. takes the weighted average of the values in each bin           #
    #       7. appends the weighted average into a list                       #
    #       8. Returns the weighted average of each bin                       #
    ###########################################################################

    # Ensure input is numpy array
    x = np.asarray(x)
    y = np.asarray(y)
    dy = np.asarray(dy)

    # Sort by x
    order = np.argsort(x)
    x, y, dy = x[order], y[order], dy[order]

    min_x = x.min()
    max_x = x.max()
    num_bins = max(1, int(np.ceil((max_x - min_x) / bin_width)))
    if num_bins < 1:
        num_bins = 1

    bins = np.linspace(min_x, max_x, num_bins + 1)
    inds = np.digitize(x, bins, right=True)

    avg_x = []
    avg_y = []
    avg_dy = []
    for i in range(1, len(bins)):
        mask = inds == i
        if np.any(mask) and np.sum(mask) >= min_points_per_bin:
            weights = 1 / dy[mask] ** 2
            # weighted mean for x and y
            wx = np.average(x[mask], weights=weights)
            wy = np.average(y[mask], weights=weights)
            wdy = 1 / np.sqrt(np.sum(weights))
            avg_x.append(wx)
            avg_y.append(wy)
            avg_dy.append(wdy)
    return np.array(avg_x), np.array(avg_y), np.array(avg_dy)

##########################################################################################
def plot_v2_fit(data_dict, star, line_spf=None, ldc_band=None, eq_text=False,
                datasets_to_plot=None, plot_ldmodel=False, plot_udmodel=False,
                to_bin=None, title=None, set_axis=None, savefig=None, show=True):
    ###########################################################################
    # Function: plot_v2_fit                                                   #
    # Inputs: data_dict -> dict of InterferometryData objects,                #
    #                    e.g. {'pavo': pavo_obj, ...}                         #
    #         star-> star object with .theta and .ldc* attributes             #
    #                (ldcR, ldcK, etc.), and .V2(line_spf, theta, ldc)        #
    #         line_spf -> x values for model curve                            #
    #         ldc_band -> string (e.g. "ldcR", "ldcK") for which LDC          #
    #                     coefficient to use                                  #
    #         eq_text -> optional string for annotation                       #
    #         datasets_to_plot-> list of keys in data_dict to plot            #
    #                            (default: all)                               #
    #         plot_ldmodel-> bool, whether to plot the ld model curve         #
    #         plot_udmodel -> bool, whether to plot the ud model curve        #
    #         to_bin -> list of kets in data_dict to bin                      #
    #         set_axis -> sets the axis limits                                #
    #         title -> allows user to set a plot title                        #
    #         savefig-> filename to save, if desired                          #
    #         show-> whether to plt.show()                                    #
    # Outputs: the plot                                                       #
    # What it does:                                                           #
    #        1. Initializes plotting parameters                               #
    #        2. Checks to see what datasets the user wants plotted            #
    #        3. Defines dictionaries for each instrument and the marker,      #
    #           color, label, and alpha value for each one                    #
    #        4. Checks to see if set_axis has been assigned. If it has,       #
    #           sets the axis limits and adjusts line_spf accordingly. If     #
    #           not, sets line_spf to be max(spf) with padding                #
    #        Starts with the top plot                                         #
    #        5. For each data set, sets the keys for the color, marker, label #
    #           and alpha value                                               #
    #        6. Checks to see if the to_bin has been set                      #
    #        7. if to_bin has been set, plots the unbinned data for each      #
    #           data set, and then bins the data sets indicated then plots    #
    #           those                                                         #
    #        8. If to_bin has not been set, it plots the unbinned data        #
    #        9. For the model, if plot_ldmodel is set, pulls the ldtheta,     #
    #           error on the theta, and the ldcs and calculates the fits for  #
    #           the relevant filter                                           #
    #       10. Plots the model for the filter indicated                      #
    #       11. If eq_text is set, annotates the plot with the theta val      #
    #       12. If plot_udmodel is set, pulls the udtheta and error, and      #
    #           calculates the UD model and plots it                          #
    #       13. If eq_text is set, annotates the plot the theta val           #
    #       14. Checks to see how many datasets are being plotted. If more    #
    #           than 1, sets the legend. If not, does not plot legend.        #
    #       Bottom plot:                                                      #
    #       15. For each data set, it pulls the respective keys for the       #
    #           color, label, alpha, and marker for each instrument           #
    #       16. For each data set, checks to see if the to_bin value has been #
    #           set.                                                          #
    #       17. If plot_ldmodel has been set, it calculates the residuals for #
    #           the data and the model for the filter indicated.              #
    #       18. If to_bin has been set, it calculates the residuals for the   #
    #           binned data as well.                                          #
    #       19. If plot_udmodel has been set, it calculates the residuals for #
    #           the data and the ud model.                                    #
    #       20. If to_bin has been set, it calcualtes the residuals for the   #
    #           binned data as well.                                          #
    #       21. Plots the residuals for the unbinned (and binned if set)      #
    #       22. Saves fig if save_fig has been set                            #
    #       23. Shows fig if show has been set.                               #
    ###########################################################################
    plt.rcParams.update({'font.size': 18})
    plt.rcParams['xtick.direction'] = 'in'
    plt.rcParams['ytick.direction'] = 'in'
    f, (a0, a1) = plt.subplots(2, 1, gridspec_kw={'height_ratios': [10, 3]}, sharex=True)

    if datasets_to_plot is None:
        datasets_to_plot = list(data_dict.keys())

    color_map = {
        'pavo': '#02ccfe',
        'classic': '#738595',
        'vega': '#5edc1f',
        'mircx': '#efc8ff',
        'mystic': '#ffcfdc',
        'spica': '#ff964f'
    }
    binned_color_map = {
        'pavo': '#030aa7',
        'classic': '#000000',
        'vega': '#028f1e',
        'mircx': '#7e1e9c',
        'mystic': '#ff028d',
        'spica': '#fe4b03'
    }
    marker_map = {
        'pavo': '.',
        'classic': 's',
        'vega': '^',
        'mircx': 'v',
        'mystic': '*',
        'spica': 'D'
    }
    label_map = {
        'pavo': r'$\rm PAVO$',
        'classic': r'$\rm Classic$',
        'vega': r'$\rm VEGA$',
        'mircx': r'$\rm MIRC-X$',
        'mystic': r'$\rm MYSTIC$',
        'spica': r'$\rm SPICA$'
    }
    binlabel_map = {
        'pavo': r'$\rm PAVO~(binned)$',
        'classic': r'$\rm Classic~(binned)$',
        'vega': r'$\rm VEGA~(binned)$',
        'mircx': r'$\rm MIRC-X~(binned)$',
        'mystic': r'$\rm MYSTIC~(binned)$',
        'spica': r'$\rm SPICA~(binned)$'
    }
    alpha_map = {
        'pavo': 0.15,
        'classic': 0.5,
        'vega': 0.5,
        'mircx': 0.15,
        'mystic': 0.15,
        'spica': 0.15
    }

    if set_axis and line_spf is None:
        xmin = set_axis[0]
        xmax = set_axis[1]
        ymin = set_axis[2]
        ymax = set_axis[3]
        a0.set_xlim(xmin, xmax)
        a0.set_ylim(ymin, ymax)
        line_spf = np.linspace(0.00001, xmax, 1000)
    else:
        all_spf = []
        for key in datasets_to_plot:
            data = data_dict[key]
            spf = np.array(data.B) / np.array(data.Wave)
            all_spf.extend(spf)
        all_spf = np.array(all_spf)
        max_spf = np.max(all_spf)
        line_spf = np.linspace(0.00001, max_spf * 1.1, 1000)  # slight padding

    # --- Top: V2 ---
    for key in datasets_to_plot:
        data = data_dict[key]
        color = color_map.get(key, None)
        bin_color = binned_color_map.get(key, None)
        bin_label = binlabel_map.get(key, None)
        marker = marker_map.get(key, '.')
        label = label_map.get(key, key.capitalize())
        alpha = alpha_map.get(key, 0.5)
        spf = np.array(data.B) / np.array(data.Wave)

        is_binned = to_bin and key in to_bin
        # Always plot both, but only one gets the label
        if is_binned:
            # Plot unbinned points, no label
            a0.plot(spf, data.V2, linestyle='None', marker=marker, markersize=3, color=color, alpha=alpha)
            a0.errorbar(spf, data.V2, yerr=abs(data.dV2), fmt=marker, markersize=3, linestyle='None', linewidth=0.5,
                        color=color,
                        capsize=3, alpha=alpha)
            # Plot binned points, with label
            binned_spf, binned_v2, binned_dv2 = bin_data(spf, data.V2, data.dV2)
            a0.plot(binned_spf, binned_v2, linestyle='None', marker=marker, markersize=6, color=bin_color,
                    label=bin_label)
            a0.errorbar(binned_spf, binned_v2, yerr=abs(binned_dv2), fmt=marker, linestyle='None', markersize=6,
                        color=bin_color,
                        capsize=3)
        else:
            # Plot unbinned points, with label
            a0.plot(spf, data.V2, linestyle='None', marker=marker, markersize=6, color=color, alpha=alpha, label=label)
            a0.errorbar(spf, data.V2, yerr=abs(data.dV2), fmt=marker, markersize=6, linestyle='None', linewidth=0.5,
                        color=color,
                        capsize=3, alpha=alpha)
    # --- Model ---
    if plot_ldmodel:
        ldc_value = getattr(star, ldc_band, None)
        theta = getattr(star, "ldtheta", None)
        dtheta = getattr(star, "ldtheta_err", None)
        if ldc_value is not None and theta is not None:
            model_label = fr"$ \rm Model ({ldc_band.replace('ldc_', '').upper()})$"
            a0.plot(line_spf, V2(line_spf, theta, ldc_value), '--', color='black', label=model_label)
            if eq_text:
                eq1 = fr"$\theta_{{\rm LD}} = {round(theta, 3):.3f} \pm {round(dtheta, 3):.3f} \rm ~mas$"
                a0.text(0.05, 0.05, eq1, transform=a0.transAxes, color='black', fontsize=15)

        else:
            print(f"Warning: {ldc_band} or ldtheta not present for star, skipping model plot.")

    if plot_udmodel:
        theta = getattr(star, "udtheta", None)
        dtheta = getattr(star, "udtheta_err", None)
        if theta is not None:
            model_label = fr"$\rm Uniform~Disk~Model$"
            a0.plot(line_spf, UDV2(line_spf, theta), '--', color='black', label=model_label)
            if eq_text:
                eq1 = fr"$\theta_{{\rm UD}} = {round(theta, 3):.3f} \pm {round(dtheta, 3):.3f} \rm ~mas$"
                a0.text(0.05, 0.05, eq1, transform=a0.transAxes, color='black', fontsize=15)
        else:
            print(f"Warning: udtheta not present for star, skipping model plot.")

    if len(datasets_to_plot) > 1:
        a0.legend(fontsize=12)

    a0.set_ylabel(r'$V^2$', labelpad=17)
    a0.tick_params(axis='x', labelbottom=False)
    a0.xaxis.set_minor_locator(AutoMinorLocator())
    a0.yaxis.set_minor_locator(AutoMinorLocator())
    a0.set_title(title)

    # --- Bottom panel: Residuals ---
    for key in datasets_to_plot:
        data = data_dict[key]
        color = color_map.get(key, None)
        bin_color = binned_color_map.get(key, None)
        marker = marker_map.get(key, '.')
        alpha = alpha_map.get(key, 0.5)
        spf = np.array(data.B) / np.array(data.Wave)

        is_binned = to_bin and key in to_bin  # e.g. to_bin = ['pavo']

        # --- Model and Residuals for Unbinned ---
        if plot_ldmodel and ldc_value is not None and theta is not None:
            model_v2 = V2(spf, theta, ldc_value)
            residuals = np.array(data.V2) - model_v2
            a1.plot(spf, residuals, linestyle='None', marker=marker, markersize=3, color=color, alpha=alpha)
            a1.errorbar(spf, residuals, yerr=abs(data.dV2), fmt=marker, markersize=3, linestyle='None', linewidth=0.5,
                        color=color,
                        capsize=3, alpha=alpha)

            # --- Model and Residuals for Binned ---
            if is_binned:
                binned_spf, binned_v2, binned_dv2 = bin_data(spf, data.V2, data.dV2)
                model_binv2 = V2(binned_spf, theta, ldc_value)
                binned_res = binned_v2 - model_binv2
                a1.plot(binned_spf, binned_res, linestyle='None', marker=marker, markersize=6, color=bin_color)
                a1.errorbar(binned_spf, binned_res, yerr=abs(binned_dv2), fmt=marker, linestyle='None', markersize=6,
                            color=bin_color, capsize=3)

        # --- (Repeat similar for UD model if desired) ---
        if plot_udmodel and theta is not None:
            model_udv2 = UDV2(spf, theta)
            ud_res = np.array(data.V2) - model_udv2
            a1.plot(spf, ud_res, linestyle='None', marker=marker, markersize=3, color=color, alpha=alpha)
            a1.errorbar(spf, ud_res, yerr=abs(data.dV2), fmt=marker, markersize=3, linestyle='None', linewidth=0.5,
                        color=color, capsize=5,
                        alpha=alpha)

            if is_binned:
                binned_spf, binned_v2, binned_dv2 = bin_data(spf, data.V2, data.dV2)
                model_binudv2 = UDV2(binned_spf, theta)
                binned_udres = binned_v2 - model_binudv2
                a1.plot(binned_spf, binned_udres, linestyle='None', marker=marker, markersize=6, color=bin_color)
                a1.errorbar(binned_spf, binned_udres, yerr=abs(binned_dv2), fmt=marker, linestyle='None', markersize=6,
                            color=bin_color, capsize=3)

    a1.axhline(y=0, color='black', linestyle='--')
    a1.set_ylabel(r'$\rm Residual$', labelpad=3)
    plt.yticks([-.25, 0, 0.25])
    a1.set_ylim([-0.35, 0.35])
    a1.xaxis.set_minor_locator(AutoMinorLocator())
    a1.yaxis.set_minor_locator(AutoMinorLocator())

    plt.subplots_adjust(wspace=0, hspace=0)
    plt.xlabel(r'$\rm Spatial$ $\rm frequency$ [$\rm rad^{-1}$]')

    if savefig:
        f.savefig(savefig, bbox_inches='tight')
    if show:
        plt.show()
    return f, (a0, a1)


import numpy as np
import pandas as pd
from scipy.interpolate import LinearNDInterpolator
from radpy.config import rppath, rapath, kppath, kapath, hppath, hapath, jppath, japath, rpfpath, rafpath, kpfpath, kafpath, hpfpath, hafpath, jpfpath, jafpath

def create_ldc_functions(df):
    ############################################################################
    # Function: create_ldc_function                                            #
    # Inputs: df -> the Claret data tables                                     #
    # Outputs: ldc_func -> the function to interpolate the ldc for the filter  #
    # What it does:                                                            #
    #      1. Drops any nans in the table                                      #
    #      2. If the model is the Phoenix models, the interpolator only takes  #
    #         the teff and logg and returns the mus.                           #
    #      3. If the model is the Atlas models, the interpolator takes the     #
    #         teff, logg, and metallicity, and returns the mus.                #
    #      4. Returns the interpolated function                                #
    ############################################################################
    df_clean = df.dropna()
    if df_clean['Mod'][0] == 'P':
        ldc_func = LinearNDInterpolator((df_clean['Teff'], df_clean['logg']), df_clean['u'])
        return ldc_func
    else:
        points = np.stack([df_clean['Teff'], df_clean['logg'], df_clean['Z']], axis = -1)
        ldc_func = LinearNDInterpolator(points, df_clean['u'])
        return ldc_func


def ldc_calc(teff, logg, feh, filt, method, verbose = False):
    ###########################################################################
    # Function: ldc_calc                                                      #
    # Inputs: teff -> the effective temperature                               #
    #         logg -> the surface gravity                                     #
    #         feh -> the metallicity                                          #
    #         filt -> the filter for the LDC                                  #
    #         method -> the method for finding the LDC                        #
    #         verbose -> if set to true, prints the print statements          #
    #                    defaults to not print                                #
    # Outputs: the limb-darkening coefficient                                 #
    # What it does:                                                           #
    #       1. returns the lower-case value for the filter to make sure there #
    #          is no case mismatch                                            #
    #       2. defines a dictionary for the Phoenix functions                 #
    #          Phoenix is used if the temps are below 3500 K                  #
    #       3. defines a dictionary for the Atlas functions                   #
    #          Atlas is used if the temps are above 3500 K                    #
    #       4. if temps are below 3500 and the logg is between 3.5 and 5,     #
    #          uses the Phoenix models to generate a limb darkening coeff.    #
    #       5. if temps are above 3500, uses the Atlas models to generate     #
    #          a limb darkening coeff.                                        #
    #       6. Returns the limb darkening coefficient.                        #
    ###########################################################################
    filt = filt.lower()
    # Functions for teff < 3500 and using Least squares
    low_teff_Lfuncs = {
        'r': RPfunc,
        'h': HPfunc,
        'k': KPfunc,
        'j': JPfunc
    }
    # Functions for teff >= 3500 and using Least squares
    high_teff_Lfuncs = {
        'r': RAfunc,
        'h': HAfunc,
        'k': KAfunc,
        'j': JAfunc
    }

    # Functions for teff < 3500 and using flux conservation
    low_teff_Ffuncs = {
        'r': RPFfunc,
        'h': HPFfunc,
        'k': KPFfunc,
        'j': JPFfunc
    }
    # Functions for teff >= 3500 and using flux conservation
    high_teff_Ffuncs = {
        'r': RAFfunc,
        'h': HAFfunc,
        'k': KAFfunc,
        'j': JAFfunc
    }
    if method.lower() == 'l':
        if teff < 3500 and 3.5 <=logg <=5:
            if verbose:
                print('Using the Phoenix models and Least Squares method.')
            func = low_teff_Lfuncs.get(filt)
            if func is None:
                raise ValueError(f"Unknown filter: {filt}")
            mu = func([teff, logg])[0]
        else:
            if verbose:
                print("Using the Atlas models and Least Squares method.")
            func = high_teff_Lfuncs.get(filt)
            if func is None:
                raise ValueError(f"Unknown filter: {filt}")
            mu = func([teff, logg, feh])[0]
        return mu
    if method.lower() == 'f':
        if teff < 3500 and 3.5 <= logg <= 5:
            if verbose:
                print('Using the Phoenix models and flux conservation method.')
            func = low_teff_Ffuncs.get(filt)
            if func is None:
                raise ValueError(f"Unknown filter: {filt}")
            mu = func([teff, logg])[0]
        else:
            if verbose:
                print("Using the Atlas models and flux conservation method.")
            func = high_teff_Ffuncs.get(filt)
            if func is None:
                raise ValueError(f"Unknown filter: {filt}")
            mu = func([teff, logg, feh])[0]
        return mu

ldc_RP = pd.read_csv(rppath)
ldc_RA = pd.read_csv(rapath)
ldc_KP = pd.read_csv(kppath)
ldc_KA = pd.read_csv(kapath)
ldc_HP = pd.read_csv(hppath)
ldc_HA = pd.read_csv(hapath)
ldc_JP = pd.read_csv(jppath)
ldc_JA = pd.read_csv(japath)

ldc_RPF = pd.read_csv(rpfpath)
ldc_RAF = pd.read_csv(rafpath)
ldc_KPF = pd.read_csv(kpfpath)
ldc_KAF = pd.read_csv(kafpath)
ldc_HPF = pd.read_csv(hpfpath)
ldc_HAF = pd.read_csv(hafpath)
ldc_JPF = pd.read_csv(jpfpath)
ldc_JAF = pd.read_csv(jafpath)

RPfunc = create_ldc_functions(ldc_RP)
RAfunc = create_ldc_functions(ldc_RA)
HPfunc = create_ldc_functions(ldc_HP)
HAfunc = create_ldc_functions(ldc_HA)
KPfunc = create_ldc_functions(ldc_KP)
KAfunc = create_ldc_functions(ldc_KA)
JPfunc = create_ldc_functions(ldc_JP)
JAfunc = create_ldc_functions(ldc_JA)

RPFfunc = create_ldc_functions(ldc_RPF)
RAFfunc = create_ldc_functions(ldc_RAF)
HPFfunc = create_ldc_functions(ldc_HPF)
HAFfunc = create_ldc_functions(ldc_HAF)
KPFfunc = create_ldc_functions(ldc_KPF)
KAFfunc = create_ldc_functions(ldc_KAF)
JPFfunc = create_ldc_functions(ldc_JPF)
JAFfunc = create_ldc_functions(ldc_JAF)


import os

raddir = os.path.dirname(os.path.realpath(__file__))
datadir = os.path.join(raddir, 'data')
# All Least squares method
hapath = os.path.join(datadir, 'ldcHA.csv')
hppath = os.path.join(datadir, 'ldcHP.csv')
japath = os.path.join(datadir, 'ldcJA.csv')
jppath = os.path.join(datadir, 'ldcJP.csv')
kapath = os.path.join(datadir, 'ldcKA.csv')
kppath = os.path.join(datadir, 'ldcKP.csv')
rapath = os.path.join(datadir, 'ldcRA.csv')
rppath = os.path.join(datadir, 'ldcRP.csv')

# All flux conservation method
hafpath = os.path.join(datadir, 'ldcHAF.csv')
hpfpath = os.path.join(datadir, 'ldcHPF.csv')
jafpath = os.path.join(datadir, 'ldcJAF.csv')
jpfpath = os.path.join(datadir, 'ldcJPF.csv')
kafpath = os.path.join(datadir, 'ldcKAF.csv')
kpfpath = os.path.join(datadir, 'ldcKPF.csv')
rafpath = os.path.join(datadir, 'ldcRAF.csv')
rpfpath = os.path.join(datadir, 'ldcRPF.csv')


classicpath = os.path.join(datadir, 'ClassicData.csv')
pavopath = os.path.join(datadir, 'PAVOdata.csv')
vegapath = os.path.join(datadir, 'Vegadata.csv')




from radpy.stellar import *
from radpy.datareadandformat import *
from radpy.plotting import plot_v2_fit
from radpy.LDfitting import initial_LDfit, run_LDfit
from radpy.config import vegapath, classicpath, pavopath
from radpy.UDfitting import initial_UDfit, run_udmcbs_fit, udfit_values


#Reading in the data files first
filenamec = classicpath
filenamep = pavopath
filenamev = vegapath

datav, num_brack_v = filename_extension(filenamev, 'V')
datac, num_brack_c = filename_extension(filenamec, 'C')
datap, num_brack_p = filename_extension(filenamep, 'P')

pavo_data = PavoData(datap)
vega_data = VegaData(datav)
classic_data = ClassicData(datac)

b, v2, dv2, wave, band, brack, inst = combined(pavo_data.make_df(), classic_data.make_df(), vega_data.make_df())
spf = b/wave

#Stellar parameters
p = 152.864
dp = 0.0494
f = 21.751
df = 0.585
logg = 4.5
dlogg = 0.1
m = 0.09
dm = 0.08
#distances("HD 219134", verbose = True)
D, dD = distances('HD 219134', verbose = True)

star = StellarParams()
star.fbol = f
star.fbol_err = df
star.logg = logg
star.logg_err = dlogg
star.feh = m
star.feh_err = dm
star.dist = D
star.dist_err = dD
star.plx = p
star.plx_err = dp

print("Stellar parameters:")
print(star)

theta1, dtheta1, chisqr1 = initial_UDfit(spf, v2, dv2, 0.4, star, verbose = False)
#initial_UDfit(spf, v2, dv2, 0.4, star, verbose = False)
theta2, dtheta2, chisqr2 = initial_LDfit(spf, v2, dv2, star, 'R', verbose = True)
#initial_LDfit(spf, v2, dv2, star, 'R', verbose = True)

print("Initial fitting results:")
print(star)

results = run_udmcbs_fit(10, 10, datasets = [pavo_data, vega_data, classic_data], stellar_params = star)
udfit_values(spf, v2, dv2, results, stellar_params = star, verbose = True)

thetaf, dthetaf, tf, dtf, ldcsf, chisf = run_LDfit(2, 2, ogdata = [spf, v2, dv2], datasets = [pavo_data, vega_data, classic_data], stellar_params = star, verbose = True)

print("Final fitting results:")
print(star)

print("Final stellar parameters:")
calc_star_params(star, verbose = True)
print(star)


data_dict = {'pavo':pavo_data, 'vega':vega_data, 'classic':classic_data}
plot_v2_fit(
    data_dict=data_dict,
    star=star,          # Your Star instance with theta, ldcR, etc., and .V2 method
    line_spf=np.linspace(0.00001, 2.5e8, 1000),
    set_axis = [0, 2.5e8, -0.05,1.1],
    ldc_band='ldc_R',        # or 'ldcR', 'ldcH', etc.
    datasets_to_plot=['pavo', 'vega', 'classic'],
    plot_udmodel = True,
    to_bin = ['pavo'],
    eq_text=True,
    show = True
)






import pandas as pd
import numpy as np
import astropy.io.fits as fits

def check_multidatasets(fitsfile, verbose = False):
    count = 0
    for i, hdu in enumerate(fitsfile):
        name = 'OI_VIS2'
        if hdu.name == name:
            if hdu.ver:
                count +=1
    if verbose:
        print("Nights:", count)
    return count

def unpack_multidatasets(fitsfile, count, verbose = False):
    if count > 1:
        v2_list = []
        dv2_list = []
        uc_list = []
        vc_list = []
        mjd_list = []
        time_list = []
        wl_list = []
        band_list = []
        if verbose:
            print("Multiple nights")
        for i in range(count):
            if verbose:
                print("Version:", i + 1)
            v2 = fitsfile["OI_VIS2", i + 1].data["VIS2DATA"]
            dv2 = fitsfile["OI_VIS2", i + 1].data["VIS2ERR"]
            ucoord = fitsfile["OI_VIS2", i + 1].data["UCOORD"]
            vcoord = fitsfile["OI_VIS2", i + 1].data["VCOORD"]
            mjd = fitsfile["OI_VIS2", i + 1].data["MJD"]
            time = fitsfile["OI_VIS2", i + 1].data["TIME"]
            wl = fitsfile["OI_WAVELENGTH", i + 1].data["EFF_WAVE"]
            band = fitsfile["OI_WAVELENGTH", i + 1].data["EFF_BAND"]

            wll = [wl.tolist() for _ in range(len(v2))]
            bandl = [band.tolist() for _ in range(len(v2))]

            v2_list.append(v2)
            dv2_list.append(dv2)
            uc_list.append(ucoord)
            vc_list.append(vcoord)
            mjd_list.append(mjd)
            time_list.append(time)
            wl_list.append(wll)
            band_list.append(bandl)
        pd.set_option('display.float_format', '{:.12f}'.format)
        df = pd.DataFrame({'MJD': mjd_list, 'Time': time_list, 'V2': v2_list, 'V2_err': dv2_list,
                           'Eff_wave[m]': wl_list, 'Eff_band[m]': band_list, 'UCOORD[m]': uc_list,
                           'VCOORD[m]': vc_list})
        df_e = df.apply(pd.Series.explode)
        return df_e
    else:
        if verbose:
            print("Only one night.")
        v2 = fitsfile["OI_VIS2"].data["VIS2DATA"]
        dv2 = fitsfile["OI_VIS2"].data["VIS2ERR"]
        ucoord = fitsfile["OI_VIS2"].data["UCOORD"]
        vcoord = fitsfile["OI_VIS2"].data["VCOORD"]
        mjd = fitsfile["OI_VIS2"].data["MJD"]
        time = fitsfile["OI_VIS2"].data["TIME"]
        wl = fitsfile["OI_WAVELENGTH"].data["EFF_WAVE"]
        band = fitsfile["OI_WAVELENGTH"].data["EFF_BAND"]

        wl_list = [wl.tolist() for _ in range(len(v2))]
        band_list = [band.tolist() for _ in range(len(v2))]

        pd.set_option('display.float_format', '{:.12f}'.format)
        df = pd.DataFrame({'MJD': mjd, 'Time': time, 'V2': v2.tolist(), 'V2_err': dv2.tolist(),
                           'Eff_wave[m]': wl_list, 'Eff_band[m]': band_list, 'UCOORD[m]': ucoord, 'VCOORD[m]': vcoord})

        return df


def oifits_to_pandas(filename, inst_name):
    # Reads in an oifits file and converts into a pandas dataframe with the necessary information needed for RADPy
    # Extracts the following data:
    #   V2, dv2, ucoord, vcoord, mjd, time, effective wavelength, and effective bandwidth
    # Converts the wavelengths and bandwidth arrays into lists of lists to match the V2 and V2err lists
    # creates a dataframe
    # Sorts the dataframe by MJD and then assigns a bracket number based on the date groupings
    # Explodes the dataframe by extracting out each list for V2, V2_err, wavelength, and bandwidth
    # returns the exploded, sorted, and bracket labeled df

    data = fits.open(filename)
    count = check_multidatasets(data)
    new_df = unpack_multidatasets(data,count)
    sorted_df = brackets(new_df, inst_name)
    sorted_df['zipped'] = sorted_df.apply(
        lambda row: list(zip(row['V2'], row['V2_err'], row['Eff_wave[m]'], row['Eff_band[m]'])), axis=1)
    df_exploded = sorted_df.explode('zipped').reset_index(drop=True)
    df_exploded[['V2', 'V2_err', 'Eff_wave[m]', 'Eff_band[m]']] = pd.DataFrame(df_exploded['zipped'].tolist(),
                                                                               index=df_exploded.index)
    df_exploded = df_exploded.drop(columns='zipped')
    df_exploded['Instrument'] = [inst_name] * len(df_exploded)

    return df_exploded


def filename_extension(filename, inst_name, verbose=False, debug = False):
    #########################################################################
    # Function: filename_extension                                          #
    # Inputs: filename -> name of data file                                 #
    #         inst_name -> Instrument identifier                            #
    #                      C - Classic                                      #
    #                      P - PAVO                                         #
    #                      V - VEGA                                         #
    #                      M - MIRCX                                        #
    #                      MY - MYSTIC                                      #
    #                      S - SPICA                                        #
    #         verbose -> default is False, if true, allows print statements #
    # Outputs: data frame of the data with the instrument added as a column #
    # What it does:                                                         #
    #         1. Checks what format the file is in                          #
    #         If .csv:                                                      #
    #            2a. Uses pandas.read_csv to read in the file               #
    #            3a. Adds the Instrument column                             #
    #         If .txt:                                                      #
    #            2b. Opens the file and reads in the first line             #
    #            3b. Checks what delimiter the file is using                #
    #            4b. Reads in the file                                      #
    #            5b. Adds the instrument column                             #
    #         If .oifits or .fits:                                          #
    #            2c. uses the oifits_to_pandas function                     #
    #         Returns the dataframe, and number of brackets                 #
    #########################################################################

    if filename.endswith('.csv'):
        df = pd.read_csv(filename)
        df['Instrument'] = [inst_name] * len(df)
        sorted_df = brackets(df, inst_name)
        num_brackets = sorted_df['Bracket'].max()
        print('Number of brackets:', num_brackets)
        return sorted_df, num_brackets
        # return sorted_df
    elif filename.endswith('.txt'):
        header = None
        data_start = 0
        with open(filename, 'r') as f:
            lines = f.readlines()
        for i, line in enumerate(lines):
            if line.strip().startswith('#'):
                # Remove '#' and newline, then split by pipe (or whatever header delimiter you expect)
                header = [col.strip() for col in line.strip()[1:].split('|')]
                data_start = i + 1
                if debug:
                    print(f"Header detected: {header}")
                break

        # Read the data lines (skip header and comments)
        # Drop empty lines and comments
        data_lines = [l for l in lines[data_start:] if l.strip() and not l.strip().startswith('#')]

        # Save to a temp string buffer for pandas
        from io import StringIO
        data_str = ''.join(data_lines)
        df = pd.read_csv(StringIO(data_str), sep=r'\s+', header=None, engine='python')
        if header and len(header) == df.shape[1]:
            df.columns = header
        df['Instrument'] = [inst_name] * len(df)
        sorted_df = brackets(df, inst_name)
        num_brackets = sorted_df['Bracket'].max()
        if verbose:
            print('Number of brackets:', num_brackets)
        return sorted_df, num_brackets


    elif filename.endswith('.oifits') or filename.endswith('.fits'):
        df = oifits_to_pandas(filename, inst_name)
        num_brackets = df['Bracket'].max()
        print('Number of brackets:', num_brackets)
        return df, num_brackets
        # return df

    else:
        header = None
        data_start = 0
        with open(filename, "r") as f:
            lines = f.readlines()
        for i, line in enumerate(lines):
            if line.strip().startswith('#'):
                # Remove '#' and newline, then split by pipe (or whatever header delimiter you expect)
                header = [col.strip() for col in line.strip()[1:].split('|')]
                data_start = i + 1
            if debug:
                print(f"Header detected: {header}")
            break

        # Read the data lines (skip header and comments)
        # Drop empty lines and comments
        data_lines = [l for l in lines[data_start:] if l.strip() and not l.strip().startswith('#')]

        # Save to a temp string buffer for pandas
        from io import StringIO
        data_str = ''.join(data_lines)
        df = pd.read_csv(StringIO(data_str), sep=r'\s+', header=None, engine='python')
        if header and len(header) == df.shape[1]:
            df.columns = header
        df['Instrument'] = [inst_name] * len(df)
        sorted_df = brackets(df, inst_name)
        num_brackets = sorted_df['Bracket'].max()
        if verbose:
            print('Number of brackets:', num_brackets)
        return sorted_df, num_brackets

def brackets(df, instrument):
    # bracket generator
    # PAVO brackets are assigned via same baseline
    # Classic has no brackets
    # MIRCX/MYSTIC/SPICA are by MJD
    # Vega is uncertain for now but will be date im fairly certain
    if instrument == 'M' or instrument == 'm':
        pd.set_option('display.float_format', '{:.12f}'.format)
        sorted_df = df.sort_values(by='MJD')
        sorted_df['Bracket'] = sorted_df.groupby('MJD').ngroup() + 1
        return sorted_df
    if instrument == 'P' or instrument == 'p':
        pd.set_option('display.float_format', '{:.12f}'.format)
        sorted_df = df.sort_values(by='U(meters)')
        sorted_df['Bracket'] = sorted_df.groupby('U(meters)').ngroup() + 1
        return sorted_df
    if instrument == 'C' or instrument == 'c':
        pd.set_option('display.float_format', '{:.12f}'.format)
        df['Bracket'] = [1] * len(df)
        return df
    if instrument == 'S' or instrument == 's':
        pd.set_option('display.float_format', '{:.12f}'.format)
        sorted_df = df.sort_values(by='UCOORD[m]')
        sorted_df['Bracket'] = sorted_df.groupby('UCOORD[m]').ngroup() + 1
        return sorted_df
    if instrument == 'My' or instrument == 'my':
        pd.set_option('display.float_format', '{:.12f}'.format)
        sorted_df = df.sort_values(by='MJD')
        sorted_df['Bracket'] = sorted_df.groupby('MJD').ngroup() + 1
        return sorted_df
    if instrument == 'V' or instrument == 'v':
        pd.set_option('display.float_format', '{:.12f}'.format)
        sorted_df = df.sort_values(by='MJD')
        sorted_df['Bracket'] = sorted_df.groupby('MJD').ngroup() + 1
        return sorted_df


def combined(*dfs, fulldf=False):
    # combines the data into one big data frame if needed

    b = pd.concat([df['B'] for df in dfs], ignore_index=True)
    v2 = pd.concat([df['V2'] for df in dfs], ignore_index=True)
    dv2 = pd.concat([df['dV2'] for df in dfs], ignore_index=True)
    wave = pd.concat([df['Wave'] for df in dfs], ignore_index=True)
    band = pd.concat([df['Band'] for df in dfs], ignore_index=True)
    brack = pd.concat([df['Bracket'] for df in dfs], ignore_index=True)
    inst = pd.concat([df['Instrument'] for df in dfs], ignore_index=True)


    if not fulldf:
        return b, v2, dv2, wave, band, brack, inst

    if fulldf:
        return pd.DataFrame({
            'B': b, 'V2': v2, 'dV2': dv2,
            'Wave': wave, 'Band': band,
            'Bracket': brack, 'Instrument': inst})


class InterferometryData:
    def __init__(self, df, instrument_code):
        self.raw = df.copy()
        self.instrument = instrument_code.lower()
        self.cleaned = None
        self.process()

    def process(self):
        raise NotImplementedError("Subclasses must implement the .process() method")

    def make_df(self, LDC=None):
        n = len(self.B)

        if LDC is None:
            ldc_col = [None] * n
        elif np.isscalar(LDC):
            ldc_col = [LDC] * n
        elif isinstance(LDC, (list, np.ndarray, pd.Series)) and len(LDC) == n:
            ldc_col = LDC
        else:
            raise ValueError(f"LDC must be None, a scalar, or a list/array of length {n}, got {LDC}.")

        return pd.DataFrame({
            "B": self.B,
            "V2": self.V2,
            "dV2": self.dV2,
            "Wave": self.Wave,
            "LDC": ldc_col,
            "Band": self.Band,
            "Bracket": self.Bracket,
            "Instrument": [self.instrument] * len(self.B)

        })

    def make_ldmcdf(self, LDC):
        # spf = self.B / self.Wave

        return pd.DataFrame({
            # "Spf":spf,
            "B": self.B,
            "V2": self.V2,
            "dV2": self.dV2,
            "LDC": LDC,
            "Bracket": self.Bracket,
            "Instrument": [self.instrument] * len(self.V2)
        })


class PavoData(InterferometryData):
    def __init__(self, df):
        super().__init__(df, instrument_code='p')

    def process(self):
        df = self.raw.dropna(subset=['V2', 'sigma_V2'])
        self.cleaned = df

        self.V2 = df['V2']
        self.dV2 = df['sigma_V2']
        self.U = df['U(meters)']
        self.V = df['V(meters)']
        self.B = np.sqrt(self.U ** 2 + self.V ** 2)
        self.Wave = self.B / df['B/lambda']
        self.Band = pd.Series(np.full(len(df), 5e-9))  # 5 nm
        self.Bracket = df['Bracket']


class ClassicData(InterferometryData):
    def __init__(self, df):
        super().__init__(df, instrument_code='c')

    def process(self):
        df = self.raw.dropna(subset=['Vis', 'Vis_e'])  # clean NaNs
        self.cleaned = df
        v = df['Vis']
        dv = df['Vis_e']

        self.B = df['B']
        self.V2 = v ** 2
        self.dV2 = self.V2 * np.sqrt(2 * (dv / v) ** 2)
        self.Wave = pd.Series(np.full(len(self.B), 2.1329e-6))  # meters
        self.Band = pd.Series(np.full(len(self.B), 5e-9))  # meters (5 nm)
        self.Bracket = df['Bracket']


class VegaData(InterferometryData):
    def __init__(self, df):
        super().__init__(df, instrument_code='v')

    def process(self):
        df = self.raw.dropna(subset=['V2', 'sigma_sys', 'sigma_stat'])  # clean NaNs
        self.cleaned = df
        dv2_sys = df['sigma_sys']
        dv2_stat = df['sigma_stat']

        self.B = df['Baseline length']
        self.V2 = df['V2']
        self.dV2 = np.sqrt((dv2_sys ** 2) + (dv2_stat) ** 2)
        self.Wave = df['lambda'] * 1e-9
        self.Band = pd.Series(np.full(len(df), 5e-9))
        self.Bracket = df['Bracket']


class MircxData(InterferometryData):
    def __init__(self, df):
        super().__init__(df, instrument_code='m')

    def process(self):
        df = self.raw.dropna(subset=['V2', 'V2_err'])  # clean NaNs
        self.cleaned = df

        ucoord = (df['UCOORD[m]'].values).astype('float')
        vcoord = (df['VCOORD[m]'].values).astype('float')
        self.B = np.sqrt((ucoord ** 2) + (vcoord ** 2))
        self.V2 = df['V2']
        self.dV2 = df['V2_err']
        self.Wave = df['Eff_wave[m]']
        self.Band = df['Eff_band[m]']
        self.Bracket = df['Bracket']

class MysticData(InterferometryData):
    def __init__(self, df):
        super().__init__(df, instrument_code='my')

    def process(self):
        df = self.raw.dropna(subset=['V2', 'V2_err'])  # clean NaNs
        self.cleaned = df

        ucoord = (df['UCOORD[m]'].values).astype('float')
        vcoord = (df['VCOORD[m]'].values).astype('float')
        self.B = np.sqrt((ucoord ** 2) + (vcoord ** 2))
        self.V2 = df['V2']
        self.dV2 = df['V2_err']
        self.Wave = df['Eff_wave[m]']
        self.Band = df['Eff_band[m]']
        self.Bracket = df['Bracket']

class SpicaData(InterferometryData):
    def __init__(self, df):
        super().__init__(df, instrument_code='s')

    def process(self):
        df = self.raw.dropna(subset=['V2', 'V2_err'])  # clean NaNs
        self.cleaned = df

        ucoord = df['UCOORD[m]']
        vcoord = df['VCOORD[m]']
        self.B = np.sqrt((ucoord ** 2) + (vcoord ** 2))
        self.V2 = df['V2']
        self.dV2 = df['V2_err']
        self.Wave = df['Eff_wave[m]']
        self.Band = df['Eff_band[m]']
        self.Bracket = df['Bracket']





import re
import os
from radpy.stellar import *
from radpy.datareadandformat import *
from radpy.plotting import plot_v2_fit
from radpy.LDfitting import initial_LDfit, run_LDfit
from radpy.UDfitting import initial_UDfit, run_UDfit, udfit_values

def extract_id(star_name):
    #############################################
    # Function: extract_id                      #
    # Inputs: star_name -> name of the star     #
    # Outputs: just the numbers of the star ID  #
    # What it does:                             #
    #     1. Takes in the star name             #
    #     2. Searches the string for one or     #
    #        more digits.                       #
    #     3. If it found some, returns the      #
    #        exact sequence of digits.          #
    #     4. If none are found, returns None    #
    #############################################
    match = re.search(r'\d+', star_name)
    return match.group(0) if match else None


def find_files_for_star(star_id, data_dir):
    ####################################################
    # Function: find_files_for_star                    #
    # Inputs: star_id -> star id numbers               #
    #         data_dir -> data directory               #
    # Outputs: matches -> list of files matching to    #
    #                     the star ID                  #
    # What it does:                                    #
    #      1. sets the ignore extension variable for   #
    #         any image extension                      #
    #      2. loops through the files in the data      #
    #         directory                                #
    #      3. Searches for the star id in the file     #
    #         name.                                    #
    #      4. Checks the file extension on each file   #
    #      5. If file is not in the ignore list,       #
    #         appends file to the matches list         #
    #      6. Returns the file list                    #
    ####################################################
    ignore_ext = {'.jpg', '.jpeg', '.png', '.pdf', '.eps', '.gif', '.tif', '.tiff', '.bmp'}
    pattern = re.compile(rf'_{star_id}\b')  # underscore, then star_id, then word boundary
    matches = []
    for root, dirs, files in os.walk(data_dir):
        for file in files:
            ext = os.path.splitext(file)[1].lower()
            if pattern.search(file) and ext not in ignore_ext:
                matches.append(os.path.join(root, file))
    return matches

def extract_instrument_from_filename(filename):
    ####################################################
    # Function: extract_instrument_from_filename       #
    # Inputs: filename -> filename                     #
    # Outputs: key -> instrument key                   #
    # What it does:                                    #
    #      1. Creates instrument dictionary with keys  #
    #      2. Pulls instrument name from file name     #
    #      3. Checks for filename and matches the key  #
    #      4. As a fall back, returns the first letter #
    #         of the filename                          #
    #      5. If not found, returns "UNKNOWN"          #
    ####################################################
    instruments = {'C': 'classic', 'P': 'pavo', 'V': 'vega', 'M': 'mircx', 'MY': 'mystic', 'S': 'spica'}
    name = os.path.basename(filename).upper()
    # Try to detect instrument by filename
    if "PAVO" in name: return "P"
    if "VEGA" in name: return "V"
    if "CLASSIC" in name or "CLIMB" in name: return "C"
    if "MIRCX" in name: return "M"
    if "MYSTIC" in name: return "MY"
    if "SPICA" in name: return "S"
    # Fallback: use first letter prefix
    for key in instruments:
        if name.startswith(key):
            return key
    return "UNKNOWN"

def save_plot(fig, out_dir, star_id, plot_type, extension):
    ##########################################################
    # Function: save_plot                                    #
    # Inputs: fig -> the figure                              #
    #         out_dir -> output directory                    #
    #         star_id -> name of star                        #
    #         plot_type -> type of model being plotted       #
    #         extension -> type of file extension            #
    # Outputs: saves figure                                  #
    # What it does:                                          #
    #      1. Checks to see if output directory has been     #
    #         made. If not, creates it.                      #
    #      2. Based on file extension, saves figure to       #
    #         output directory                               #
    ##########################################################
    os.makedirs(out_dir, exist_ok=True)
    if extension == '.png':
        fig.savefig(os.path.join(out_dir, f"{star_id}_{plot_type}.png"), bbox_inches = 'tight', dpi = 200)
    if extension == '.jpg':
        fig.savefig(os.path.join(out_dir, f"{star_id}_{plot_type}.jpg"), bbox_inches='tight', dpi = 200)
    if extension == '.eps':
        fig.savefig(os.path.join(out_dir, f"{star_id}_{plot_type}.eps"), bbox_inches='tight', dpi = 200)
    if extension == '.pdf':
        fig.savefig(os.path.join(out_dir, f"{star_id}_{plot_type}.pdf"), bbox_inches='tight', dpi = 200)

def write_latex_table(df, out_file):
    #####################################################
    # Function: write_latex_table                       #
    # Inputs: df -> dataframe to write to table         #
    #         out_file -> output tex file               #
    # Outputs: writes to a latex file                   #
    # What it does:                                     #
    #       1. Opens the out_file                       #
    #       2. Writes to the file                       #
    #####################################################
    with open(out_file, "w") as f:
        f.write(df.to_latex(index=False, float_format="%.4f"))

def get_stellar_params(file_path):
    ######################################################
    # Function: get_stellar_params                       #
    # Inputs: file_path -> path to file                  #
    # Outputs: star_names -> names of the stars          #
    #          star_params -> dictionary of stellar      #
    #                         params                     #
    # What it does:                                      #
    #     1. reads in the file                           #
    #     2. extracts the star names                     #
    #     3. For each star name, extracts the stellar    #
    #        parameters and adds it to the dictionary    #
    #     4. Returns the star names and the dictionary   #
    ######################################################
    df = pd.read_csv(file_path, sep='\t')
    star_names = df['Star'].tolist()
    params_dict = {}
    for _, row in df.iterrows():
        params_dict[row['Star']] = {
            'fbol': row['fbol'],
            'fbol_err': row['dfbol'],
            'logg': row['logg'],
            'logg_err': row['dlogg'],
            'feh': row['feh'],
            'feh_err': row['dfeh'],
        }
    return star_names, params_dict


def converttoIDobj(dfs, verbose=False):
    ####################################################
    # Function: converttoIDobj                         #
    # Inputs: dfs -> dictionaries                      #
    # Outputs: wrapped_data -> instrument class        #
    # What it does:                                    #
    #        1. Creates instrument class map           #
    #        2. Loops through the dictionaries and     #
    #           sorts the files by instrument          #
    #        3. Converts the data file to the instru-  #
    #           ment class                             #
    #        4. Returns the instrument class object    #
    ####################################################
    # Allows for dynamic wrapping of the different instrument classes
    instrument_class_map = {
        "P": PavoData,
        "V": VegaData,
        "C": ClassicData,
        "M": MircxData,
        "My": MysticData,
        "S": SpicaData
    }
    wrapped_data = {}
    for inst_code, df in dfs.items():
        data_class = instrument_class_map.get(inst_code)
        if data_class:
            try:
                wrapped_data[inst_code] = data_class(df)
            except Exception as e:
                print(f"Failed to wrap data for {inst_code}: {e}")
        else:
            print(f"No class found for '{inst_code}'")
    return wrapped_data


def data_dict_plotting(wrapped_data):
    #####################################################
    # Function: data_dict_plotting                      #
    # Inputs: wrapped_data -> Instrument Class object   #
    # Outputs: data_dict -> data dictionary             #
    # What it does:                                     #
    #       1. Creates instrument map                   #
    #       2. Creates data dictionary for each data    #
    #          file                                     #
    #       3. Returns dictionary                       #
    #####################################################
    instrument_name_map = {
        "P": "pavo",
        "V": "vega",
        "C": "classic",
        "M": "mircx",
        "My": "mystic",
        "S": "spica"
    }
    data_dict = {
        instrument_name_map[code]: data
        for code, data in wrapped_data.items()
        if code in instrument_name_map
    }

    return data_dict

def format_catalog_name(name):
    ########################################################
    # Function: format_catalog_name                        #
    # Inputs: name -> star name                            #
    # Outputs: latex formatted name                        #
    # What it does:                                        #
    #       1. Checks to see if the name is already in     #
    #          correct format                              #
    #       2. Replaces the underscores in the name with   #
    #          spaces if there are underscores             #
    #       3. Inserts a space between the prefix and the  #
    #          numbers only at the first transition        #
    #       4. Replaces the spaces with '~'                #
    #       5. Returns the latex formatted name            #
    ########################################################

    if name.startswith('$\\rm') and name.endswith('$'):
        return name
    name = name.replace('_', ' ')
    name = re.sub(r'^([A-Za-z+\-]+)\s*([0-9].*)', r'\1 \2', name)
    return rf'$\rm {name.replace(" ", "~")}$'

def convert_names_to_latex(names):
    #####################################################
    # Function: convert_names_to_latex                  #
    # Inputs: names -> list of names                    #
    # Outputs: names2 -> list of latex formatted names  #
    # What it does:                                     #
    #      1. For every name in the names list, formats #
    #         the name with format_catalog_name()       #
    #      2. Returns formatted names                   #
    #####################################################
    names2 = [format_catalog_name(name) for name in names]
    return names2


def process_star(star_name, data_dir, output_dir, stellar_param_dict, latex_rows, ldc_method, mc_num=71, bs_num=71,
                 set_axis=None, image_ext=None, binned=None, ldc_band=None, v0_flag = False, verbose=True):
    ##################################################################
    # Function: process_star                                         #
    # Inputs: star_name -> name of star                              #
    #         data_dir -> data directory                             #
    #         output_dir -> output directory                         #
    #         stellar_params_dict -> dictionary with stellar params  #
    #         latex_rows -> for latex table                          #
    #         mc_num -> number of Monte Carlo iterations             #
    #                   defaults to 71                               #
    #         bs_num -> number of bootstrap iterations               #
    #                   defaults to 71                               #
    #         set_axis -> axis limits [xmin, xmax, ymin, ymax]       #
    #                     defaults to none                           #
    #         image_ext -> image extension for file                  #
    #                      Options: '.jpg', '.png', '.pdf', '.eps'   #
    #         binned -> data you want binned i.e. ['pavo']           #
    #         ldc_band -> limb darkening band you want               #
    #                     Options: 'ldc_R', 'ldc_H', 'ldc_K', 'ldc_J'#
    #         v0_flag - > allows you to fit for a scaling factor     #
    #                     V0^2, Default is false.                    #
    #         verbose -> allows print statements to screen           #
    # Outputs: None                                                  #
    # What it does:                                                  #
    #      1. Extracts the star name                                 #
    #      2. Finds the files for the star                           #
    #      3. Groups the files by instrument.                        #
    #      4. Process the files by instrument                        #
    #      5. Combines the data files per instrument.                #
    #      6. wraps the data into InterferometryData Objects         #
    #      7. Combines all data into one dataframe                   #
    #      8. Sets up the StellarParams object                       #
    #      9. Calculates the distance                                #
    #     10. Performs initial fits for uniform disk and limb-dark   #
    #     11. Performs the UD MC fit                                 #
    #     12. Calculates parameters after UD fit                     #
    #     13. Performs the LD MC fit                                 #
    #     14. Calculates final stellar parameters                    #
    #     15. Sets up plot directory                                 #
    #     16. If data are to be binned, creates a binned data dict   #
    #     17. converts the star name into latex format               #
    #     18. Plots the UD fit                                       #
    #     19. Saves the plot                                         #
    #     20. Plots the LD fit                                       #
    #     21. Saves the plot                                         #
    #     22. Appends the stellar parameters to the latex_rows       #
    ##################################################################
    star_id = extract_id(star_name)
    print("--------------------------------------------------")
    print(f"Starting processing for {star_name}")
    files = find_files_for_star(star_id, data_dir)
    if not files:
        print(f"No files found for {star_name} ({star_id})")
        return

    # Group files by instrument
    grouped_files = {}
    for file in files:
        inst = extract_instrument_from_filename(file)
        grouped_files.setdefault(inst, []).append(file)

    # Process files by instrument
    dataframes = {}
    for inst, filelist in grouped_files.items():
        dfs = []
        for file in filelist:
            try:
                df, _ = filename_extension(file, inst, verbose=verbose)
                dfs.append(df)
            except Exception as e:
                print(f"Failed to process {file} ({inst}): {e}")
        if dfs:
            # Concatenate all files for this instrument
            dataframes[inst] = pd.concat(dfs, ignore_index=True)

    # Convert to InterferometryData objects and wrap into instrument classes
    wrap_data = converttoIDobj(dataframes, verbose=False)

    # Combine all available data using RADPy's combined() utility
    combined_args = [obj.make_df() for obj in wrap_data.values()]

    b, v2, dv2, wave, band, brack, inst = combined(*combined_args)
    spf = b / wave

    # Set up stellar parameters
    star = StellarParams()
    params = stellar_param_dict.get(star_name, {})
    for param, value in params.items():
        setattr(star, param, value)
    # Calculate distance
    if params.get('plx', None) is not None:
        D, dD = distances(star_name, plx=params['plx'], dplx=params['plx_err'], verbose=verbose)
        star.dist = D
        star.dist_err = dD
    else:
        D, dD = distances(star_name, verbose=verbose)
        star.dist = D
        star.dist_err = dD
    initial_ldc_map = {
        'ldc_r': 'R',
        'ldc_k': 'K',
        'ldc_h': 'H',
        'ldc_j': 'J'
    }
    # Initial fits
    initial_UDfit(spf, v2, dv2, 1.0, star, v0_flag , verbose=verbose)
    initial_LDfit(spf, v2, dv2, star, initial_ldc_map[ldc_band], ldc_method, v0_flag, verbose=verbose)

    # Monte Carlo uniform-disk fit
    datasets = list(wrap_data.values())
    results_UD = run_UDfit(bs_num, mc_num, datasets=datasets, stellar_params=star, v0_flag=v0_flag)
    udfit_values(spf, v2, dv2, results_UD, stellar_params=star, v0_flag=v0_flag, verbose=verbose)

    # Monte Carlo limb-darkened fit
    run_LDfit(bs_num, mc_num, ogdata=[spf, v2, dv2], datasets=datasets, stellar_params=star, ldc_method = ldc_method, v0_flag = v0_flag, verbose=verbose)

    # Calculate additional stellar parameters
    calc_star_params(star, verbose=verbose)

    # Save plots
    plot_dir = os.path.join(output_dir, "plots")
    os.makedirs(plot_dir, exist_ok=True)

    data_dict = data_dict_plotting(wrap_data)
    if binned:
        bin_only = [k for k in binned if k in data_dict]
    else:
        bin_only = None

    star_title = convert_names_to_latex([star_name])
    # Uniform disk plot
    fig1, _ = plot_v2_fit(
        data_dict=data_dict,
        star=star,
        datasets_to_plot=list(data_dict.keys()),
        to_bin=bin_only,
        plot_udmodel=True,
        eq_text=True,
        set_axis=set_axis,
        title=star_title[0],
        show=False
    )
    save_plot(fig1, plot_dir, star_id, "UDfit", image_ext)

    # Limb-darkened R band plot
    fig2, _ = plot_v2_fit(
        data_dict=data_dict,
        star=star,
        datasets_to_plot=list(data_dict.keys()),
        to_bin=bin_only,
        plot_ldmodel=True,
        ldc_band=ldc_band,
        title=star_title[0],
        set_axis=set_axis,
        eq_text=True,
        show=False
    )
    save_plot(fig2, plot_dir, star_id, "LDfit", image_ext)

    # Collect results for LaTeX
    if v0_flag:
        latex_rows.append({
            "Star": star_name,
            "D (pc)": star.dist,
            r"$\Delta \rm D (pc)$": star.dist_err,
            r"$V_{0}^{2}$": star.ldv02,
            r"$\Delta V_{0}^{2}$": star.ldv02_err,
            r"$\theta_{\rm UD}$ (mas)": star.udtheta,
            r"$\Delta\theta_{\rm UD}$ (mas)": star.udtheta_err,
            r"$\theta_{\rm LD}$ (mas)": star.ldtheta,
            r"$\Delta\theta_{\rm LD}$ (mas)": star.ldtheta_err,
            r"$T_{\rm eff}$ (K)": star.teff,
            r"$\Delta T_{\rm eff}$ (K)": star.teff_err,
            r"$L_{\star} (\rm L_{\odot})$": star.lum,
            r"$\Delta L_{\star} (\rm L_{\odot})$": star.lum_err,
            r"$R_{\star} (\rm R_{\odot})$": star.rad,
            r"$\Delta R_{\star} (\rm R_{\odot})$": star.rad_err,
            r"$\mu_{\rm R}$": star.ldc_R,
            r"$\mu_{\rm K}$": star.ldc_K,
            r"$\mu_{\rm H}$": star.ldc_H,
            r"$\mu_{\rm J}$": star.ldc_J,
        })
    if not v0_flag:
        latex_rows.append({
            "Star": star_name,
            "D (pc)": star.dist,
            r"$\Delta \rm D (pc)$": star.dist_err,
            r"$\theta_{\rm UD}$ (mas)": star.udtheta,
            r"$\Delta\theta_{\rm UD}$ (mas)": star.udtheta_err,
            r"$\theta_{\rm LD}$ (mas)": star.ldtheta,
            r"$\Delta\theta_{\rm LD}$ (mas)": star.ldtheta_err,
            r"$T_{\rm eff}$ (K)": star.teff,
            r"$\Delta T_{\rm eff}$ (K)": star.teff_err,
            r"$L_{\star} (\rm L_{\odot})$": star.lum,
            r"$\Delta L_{\star} (\rm L_{\odot})$": star.lum_err,
            r"$R_{\star} (\rm R_{\odot})$": star.rad,
            r"$\Delta R_{\star} (\rm R_{\odot})$": star.rad_err,
            r"$\mu_{\rm R}$": star.ldc_R,
            r"$\mu_{\rm K}$": star.ldc_K,
            r"$\mu_{\rm H}$": star.ldc_H,
            r"$\mu_{\rm J}$": star.ldc_J,
        })

    print(f"Finished processing {star_name}")

def batch_mode(star_file, data_dir, output_dir, latex_out, ldc_method, mc_num=71, bs_num=71, set_axis = None, image_ext=None, binned=None, ldc_band=None, v0_flag = False, verbose=True):
    ######################################################
    # Function: batch_mode                               #
    # Inputs: star_file -> stellar param file            #
    #         data_dir -> data directory                 #
    #         output_dir -> output directory             #
    #         latex_out -> name of latex file            #
    #         mc_num -> number of MC iterations          #
    #                   defaults to 71                   #
    #         bs_num -> number of bootstrap iterations   #
    #                   defaults to 71                   #
    #         set_axis -> sets the axis limits           #
    #                   defaults to None (automatically  #
    #                   assigns)                         #
    #         image_ext -> sets image extension          #
    #         binned -> datasets you want binned         #
    #         ldc_band -> limb-darkening coefficent you  #
    #                     want plotted                   #
    #         v0_flag - > allows you to fit for a scale  #
    #                     factor, V0^2                   #
    #                     Default is false.              #
    #         verbose -> allows print statements         #
    # Outputs: None                                      #
    # What it does:                                      #
    #       1. Changes directory to the data directory   #
    #       2. Extracts star names and stellar params    #
    #       3. Assigns empty rows for latex_rows         #
    #       4. Initializes count iterator                #
    #       5. Loops through the star names and runs     #
    #          process_star for each                     #
    #       6. Updates loop counter                      #
    #       7. Outside loop, creates a DataFrame of the  #
    #          stellar parameters                        #
    #       8. Writes the dataframe to a latex table     #
    ######################################################
    os.chdir(data_dir)
    star_names, star_params = get_stellar_params(star_file)
    latex_rows = []
    count = 0
    for star_name in star_names:
        process_star(star_name, data_dir, output_dir, star_params, latex_rows, mc_num=mc_num, bs_num=bs_num, set_axis = set_axis,
                     image_ext=image_ext, binned=binned, ldc_band=ldc_band, ldc_method = ldc_method, v0_flag = v0_flag, verbose=verbose)
        count += 1

    latex_df = pd.DataFrame(latex_rows)
    write_latex_table(latex_df, latex_out)
    print(f"Batch complete. Fit {count} stars. Plots in {os.path.join(output_dir, 'plots')}, results in {latex_out}")